@@ -84,10 +84,14 @@ typedef enum {
 	t_process,
 	t_lock,
 	t_condition_variable,
+# ifdef ECL_SEMAPHORES
+        t_semaphore,
+# endif
 #endif
 	t_codeblock,
 	t_foreign,
 	t_frame,
+	t_weak_pointer,
 	t_end,
 	t_other,
 	t_contiguous,		/*  contiguous block  */
@@ -117,6 +121,9 @@ typedef cl_object (*cl_objectfn_fixed)();
 #define IMMEDIATE(o)		((cl_fixnum)(o) & 3)
 #define IMMEDIATE_TAG		3
 
+#define ecl_to_bool(x) ((x)!=Cnil)
+#define ecl_make_bool(x) ((x)? Ct : Cnil)
+
 /* Immediate fixnums:		*/
 #define FIXNUM_TAG		t_fixnum
 #define MAKE_FIXNUM(n)		((cl_object)(((cl_fixnum)(n) << 2) | FIXNUM_TAG))
@@ -125,6 +132,14 @@ typedef cl_object (*cl_objectfn_fixed)();
 #define	fix(obje)		(((cl_fixnum)(obje)) >> 2)
 #define FIXNUMP(o)		(IMMEDIATE(o) == t_fixnum)
 
+#define ECL_FIXNUMP(o)          (IMMEDIATE(o) == t_fixnum)
+#define ecl_fixnum_lower(a,b)   ((cl_fixnum)(a) < (cl_fixnum)(b))
+#define ecl_fixnum_greater(a,b) ((cl_fixnum)(a) > (cl_fixnum)(b))
+#define ecl_fixnum_leq(a,b)     ((cl_fixnum)(a) <= (cl_fixnum)(b))
+#define ecl_fixnum_geq(a,b)     ((cl_fixnum)(a) >= (cl_fixnum)(b))
+#define ecl_fixnum_plusp(a)     ((cl_fixnum)(a) > (cl_fixnum)MAKE_FIXNUM(0))
+#define ecl_fixnum_minusp(a)    ((cl_fixnum)(a) < (cl_fixnum)(0))
+
 /* Immediate characters:	*/
 #define CHARACTER_TAG		t_character
 #define CHARACTERP(o)		(IMMEDIATE(o) == t_character)
@@ -137,32 +152,55 @@ typedef cl_object (*cl_objectfn_fixed)();
 #define	CODE_CHAR(c)		((cl_object)(((cl_fixnum)((c & 0xff) << 2)|CHARACTER_TAG)))
 #define	CHAR_CODE(obje)		((((cl_fixnum)(obje)) >> 2) & 0xff)
 #endif
+#define ECL_CHAR_CODE_RETURN	13
+#define ECL_CHAR_CODE_NEWLINE	10
+#define ECL_CHAR_CODE_LINEFEED	10
 
 #define ECL_NUMBER_TYPE_P(t)	(t >= t_fixnum && t <= t_complex)
 #define REAL_TYPE(t)		(t >= t_fixnum && t < t_complex)
 #define ARRAY_TYPE(t)		(t >= t_array && t <= t_bitvector)
-#define ARRAYP(x)		((IMMEDIATE(x) == 0) && (x)->d.t >= t_array && (x)->d.t <= t_bitvector)
-#define VECTORP(x)		((IMMEDIATE(x) == 0) && (x)->d.t >= t_vector && (x)->d.t <= t_bitvector)
+#define ECL_ARRAYP(x)		((IMMEDIATE(x) == 0) && (x)->d.t >= t_array && (x)->d.t <= t_bitvector)
+#define ECL_VECTORP(x)		((IMMEDIATE(x) == 0) && (x)->d.t >= t_vector && (x)->d.t <= t_bitvector)
+#define ECL_BIT_VECTOR_P(x)     ((IMMEDIATE(x) == 0) && ((x)->d.t == t_bitvector))
+#ifdef ECL_UNICODE
+#define ECL_STRINGP(x)		((IMMEDIATE(x) == 0) && \
+                                 ((x)->d.t == t_base_string || (x)->d.t == t_string))
+#define ECL_EXTENDED_STRING_P(x) ((IMMEDIATE(x) == 0) && (x)->d.t == t_string)
+#else
+#define ECL_STRINGP(x)		((IMMEDIATE(x) == 0) && ((x)->d.t == t_base_string))
+#define ECL_EXTENDED_STRING_P(x) 0
+#endif
+#define ECL_BASE_STRING_P(x)	((IMMEDIATE(x) == 0) && ((x)->d.t == t_base_string))
+#define ECL_HASH_TABLE_P(x)	((IMMEDIATE(x) == 0) && ((x)->d.t == t_hashtable))
+#define ECL_BIGNUMP(x)          ((IMMEDIATE(x) == 0) && ((x)->d.t == t_bignum))
+#define ECL_COMPLEXP(x)         ((IMMEDIATE(x) == 0) && ((x)->d.t == t_complex))
+#define ECL_RANDOM_STATE_P(x)   ((IMMEDIATE(x) == 0) && ((x)->d.t == t_random))
+#define ECL_SINGLE_FLOAT_P(x)   ((IMMEDIATE(x) == 0) && ((x)->d.t == t_singlefloat))
+#define ECL_DOUBLE_FLOAT_P(x)   ((IMMEDIATE(x) == 0) && ((x)->d.t == t_doublefloat))
+#ifdef ECL_LONG_FLOAT
+#define ECL_LONG_FLOAT_P(x)     ((IMMEDIATE(x) == 0) && ((x)->d.t == t_longfloat))
+#endif
+#define ECL_PACKAGEP(x)         ((IMMEDIATE(x) == 0) && ((x)->d.t == t_package))
+#define ECL_PATHNAMEP(x)        ((IMMEDIATE(x) == 0) && ((x)->d.t == t_pathname))
+#define ECL_READTABLEP(x)       ((IMMEDIATE(x) == 0) && ((x)->d.t == t_readtable))
 
 #define HEADER			int8_t t, m, padding[2]
 #define HEADER1(field)		int8_t t, m, field, padding
 #define HEADER2(field1,field2)	int8_t t, m, field1, field2
-#define HEADER3(field1,flag2,flag3) int8_t t, m, field1; uint8_t flag2:4, flag3:4
-#define HEADER4(field1,flag2,flag3,flag4) int8_t t, m, field1; uint8_t flag2:4, flag3:2, flag4:2
 
 struct ecl_singlefloat {
 	HEADER;
 	float SFVAL;	/*  singlefloat value  */
 };
 #define	sf(obje)	(obje)->SF.SFVAL
-#define ecl_single_float(o) ((o)->singlefloat.value)
+#define ecl_single_float(o) ((o)->SF.value)
 
 struct ecl_doublefloat {
 	HEADER;
 	double DFVAL;	/*  doublefloat value  */
 };
 #define	df(obje)	(obje)->DF.DFVAL
-#define ecl_double_float(o) ((o)->doublefloat.value)
+#define ecl_double_float(o) ((o)->DF.value)
 
 #ifdef ECL_LONG_FLOAT
 struct ecl_long_float {
@@ -184,11 +222,11 @@ struct ecl_bignum {
 
 #else  /* WITH_GMP */
 
-# ifdef HAVE_LONG_LONG
+# ifdef ecl_long_long_t
      typedef long long int big_num_t;
-# else  /* HAVE_LONG_LONG */
+# else
      typedef long int big_num_t; /* would it work? */
-# endif /* HAVE_LONG_LONG */
+# endif /* ecl_long_long_t */
 
 struct ecl_bignum {
 	HEADER;
@@ -226,9 +264,11 @@ enum ecl_stype {		/*  symbol type  */
 #define ECL_UNBOUND		((cl_object)(cl_symbols+2))
 #define ECL_PROTECT_TAG		((cl_object)(cl_symbols+3))
 
+#define ECL_FLAG_MACRO        1
+#define ECLF_LAG_SPECIAL_VAR  4
+
 struct ecl_symbol {
-	HEADER4(stype, mflag, isform, dynamic);
-				/*  symbol type and whether it names a macro */
+	HEADER2(stype, dynamic);/*  symbol type, special-variable-p */
 	cl_object value;	/*  global value of the symbol  */
 				/*  Coincides with cons.car  */
 	cl_object gfdef;	/*  global function definition  */
@@ -241,6 +281,9 @@ struct ecl_symbol {
 	cl_object name;		/*  print name  */
 	cl_object hpack;	/*  home package  */
 				/*  Cnil for uninterned symbols  */
+#ifdef ECL_THREADS
+        cl_index binding;	/*  index into the bindings array  */
+#endif
 };
 #define SYM_FUN(sym)	((sym)->symbol.gfdef)
 
@@ -253,9 +296,6 @@ struct ecl_package {
 	cl_object usedby;	/*  used-by-list of packages  */
 	cl_object internal;	/*  hashtable for internal symbols  */
 	cl_object external;	/*  hashtable for external symbols  */
-#ifdef ECL_THREADS
-	pthread_mutex_t lock;	/*  thread safe packages  */
-#endif
 };
 
 /*
@@ -286,6 +326,11 @@ struct ecl_package {
  */
 
 #ifdef ECL_SMALL_CONS
+#define ECL_LISTP(x)    (IMMEDIATE(x) == t_list)
+#define ECL_CONSP(x)    (LISTP(x) && !Null(x))
+#define ECL_ATOM(x)     (Null(x) || !LISTP(x))
+#define ECL_SYMBOLP(x)  (Null(x) || ((IMMEDIATE(x) == 0) && ((x)->d.t == t_symbol)))
+
 #define LISTP(x)	(IMMEDIATE(x) == t_list)
 #define CONSP(x)	(LISTP(x) && !Null(x))
 #define ATOM(x)		(Null(x) || !LISTP(x))
@@ -303,6 +348,11 @@ struct ecl_cons {
 	cl_object cdr;		/*  cdr  */
 };
 #else
+#define ECL_LISTP(x)    (IMMEDIATE(x)? Null(x) : ((x)->d.t == t_list))
+#define ECL_CONSP(x)    ((IMMEDIATE(x) == 0) && ((x)->d.t == t_list))
+#define ECL_ATOM(x)     (IMMEDIATE(x) || ((x)->d.t != t_list))
+#define ECL_SYMBOLP(x)  (Null(x) || ((IMMEDIATE(x) == 0) && ((x)->d.t == t_symbol)))
+
 #define LISTP(x)	(IMMEDIATE(x)? Null(x) : ((x)->d.t == t_list))
 #define CONSP(x)	((IMMEDIATE(x) == 0) && ((x)->d.t == t_list))
 #define ATOM(x)		(IMMEDIATE(x) || ((x)->d.t != t_list))
@@ -334,16 +384,19 @@ struct ecl_hashtable_entry {	/*  hash table entry  */
 };
 
 struct ecl_hashtable {		/*  hash table header  */
-	HEADER2(test,lockable);
+	HEADER1(test);
 	struct ecl_hashtable_entry *data; /*  pointer to the hash table  */
 	cl_index entries;	/*  number of entries  */
 	cl_index size;		/*  hash table size  */
+	cl_index limit;		/*  hash table threshold (integer value)  */
 	cl_object rehash_size;	/*  rehash size  */
 	cl_object threshold;	/*  rehash threshold  */
 	double factor;		/*  cached value of threshold  */
 #ifdef ECL_THREADS
-	pthread_mutex_t lock;	/*  mutex to prevent race conditions  */
+	cl_object lock;		/*  mutex to prevent race conditions  */
 #endif
+	struct ecl_hashtable_entry *(*get)(cl_object, cl_object);
+	cl_object (*set)(cl_object, cl_object, cl_object);
 };
 
 typedef enum {			/*  array element type  */
@@ -356,6 +409,15 @@ typedef enum {			/*  array element type  */
 	/* Below here, list types accepted by streams (i.e. OPEN) */
 	aet_b8,			/*  byte8	     */
 	aet_i8,			/*  integer8	     */
+#ifdef ecl_uint16_t
+        aet_b16, aet_i16,
+#endif
+#ifdef ecl_uint32_t
+        aet_b32, aet_i32,
+#endif
+#ifdef ecl_uint64_t
+        aet_b64, aet_i64,
+#endif
 #ifdef ECL_UNICODE
 	aet_ch,			/*  character        */
 #endif
@@ -363,11 +425,33 @@ typedef enum {			/*  array element type  */
 	aet_last_type = aet_bc
 } cl_elttype;
 
+typedef unsigned char ecl_base_char;
+#ifdef ECL_UNICODE
+typedef int32_t ecl_character;
+#else
+typedef int ecl_character;
+#endif
+
 union ecl_array_data {
 	cl_object     *t;
-        unsigned char *ch;
+        ecl_base_char *bc;
+#ifdef ECL_UNICODE
+        ecl_character *c;
+#endif
 	uint8_t       *b8;
 	int8_t        *i8;
+#ifdef ecl_uint16_t
+        ecl_uint16_t  *b16;
+        ecl_int16_t   *i16;
+#endif
+#ifdef ecl_uint32_t
+        ecl_uint32_t  *b32;
+        ecl_int32_t   *i32;
+#endif
+#ifdef ecl_uint64_t
+        ecl_uint64_t  *b64;
+        ecl_int64_t   *i64;
+#endif
 	float         *sf;
 	double        *df;
 	cl_fixnum     *fix;
@@ -375,57 +459,61 @@ union ecl_array_data {
 	byte          *bit;
 };
 
-struct ecl_array {			/*  array header  */
+#define ECL_FLAG_HAS_FILL_POINTER 1
+#define ECL_FLAG_ADJUSTABLE       2
+#define ECL_ADJUSTABLE_ARRAY_P(x) ((x)->array.flags & ECL_FLAG_ADJUSTABLE)
+#define ECL_ARRAY_HAS_FILL_POINTER_P(x) ((x)->array.flags & ECL_FLAG_HAS_FILL_POINTER)
+
+struct ecl_array {		/*  array header  */
 				/*  adjustable flag  */
 				/*  has-fill-pointer flag  */
-	HEADER2(adjustable,rank);
+	HEADER2(elttype,flags);	/*  array element type, has fill ptr, adjustable-p */
 	cl_object displaced;	/*  displaced  */
 	cl_index dim;		/*  dimension  */
 	cl_index *dims;		/*  table of dimensions  */
 	union ecl_array_data self;	/*  pointer to the array  */
-	byte	elttype;	/*  element type  */
 	byte	offset;		/*  bitvector offset  */
+        byte    rank;		/*  rank of array = # of dimensions  */
 };
 
-struct ecl_vector {			/*  vector header  */
+struct ecl_vector {		/*  vector header  */
 				/*  adjustable flag  */
 				/*  has-fill-pointer flag  */
-	HEADER2(adjustable,hasfillp);
+	HEADER2(elttype,flags);	/*  array element type, has fill ptr, adjustable-p */
 	cl_object displaced;	/*  displaced  */
 	cl_index dim;		/*  dimension  */
 	cl_index fillp;		/*  fill pointer  */
 				/*  For simple vectors,  */
 				/*  v_fillp is equal to v_dim.  */
 	union ecl_array_data self;	/*  pointer to the vector  */
-	byte	elttype;	/*  element type  */
 	byte	offset;
 };
 
 struct ecl_base_string {	/*  string header  */
 				/*  adjustable flag  */
 				/*  has-fill-pointer flag  */
-	HEADER2(adjustable,hasfillp);
+	HEADER2(elttype,flags);	/*  array element type, has fill ptr, adjustable-p */
 	cl_object displaced;	/*  displaced  */
 	cl_index dim;       	/*  dimension  */
 				/*  string length  */
 	cl_index fillp;		/*  fill pointer  */
 				/*  For simple strings,  */
 				/*  st_fillp is equal to st_dim-1.  */
-	unsigned char *self;	/*  pointer to the string  */
+	ecl_base_char *self;	/*  pointer to the string  */
 };
 
 #ifdef ECL_UNICODE
 struct ecl_string {		/*  string header  */
 				/*  adjustable flag  */
 				/*  has-fill-pointer flag  */
-	HEADER2(adjustable,hasfillp);
+	HEADER2(elttype,flags);	/*  array element type, has fill ptr, adjustable-p */
 	cl_object displaced;	/*  displaced  */
 	cl_index dim;       	/*  dimension  */
 				/*  string length  */
 	cl_index fillp;		/*  fill pointer  */
 				/*  For simple strings,  */
 				/*  st_fillp is equal to st_dim-1.  */
-	cl_object *self;	/*  pointer to the string  */
+	ecl_character *self;	/*  pointer to the string  */
 };
 #endif
 
@@ -454,8 +542,11 @@ struct ecl_structure {		/*  structure header  */
 
 enum ecl_smmode {		/*  stream mode  */
 	smm_input,		/*  input  */
+	smm_input_file,		/*  input  */
 	smm_output,		/*  output  */
+	smm_output_file,	/*  output  */
 	smm_io,			/*  input-output  */
+	smm_io_file,		/*  input-output  */
 	smm_synonym,		/*  synonym  */
 	smm_broadcast,		/*  broadcast  */
 	smm_concatenated,	/*  concatenated  */
@@ -472,22 +563,100 @@ enum ecl_smmode {		/*  stream mode  */
 #endif
 };
 
+struct ecl_file_ops {
+	cl_index (*write_byte8)(cl_object strm, unsigned char *c, cl_index n);
+	cl_index (*read_byte8)(cl_object strm, unsigned char *c, cl_index n);
+
+	void (*write_byte)(cl_object c, cl_object strm);
+	cl_object (*read_byte)(cl_object strm);
+
+	int (*read_char)(cl_object strm);
+	int (*write_char)(cl_object strm, int c);
+	void (*unread_char)(cl_object strm, int c);
+	int (*peek_char)(cl_object strm);
+
+	cl_index (*read_vector)(cl_object strm, cl_object data, cl_index start, cl_index end);
+	cl_index (*write_vector)(cl_object strm, cl_object data, cl_index start, cl_index end);
+
+	int (*listen)(cl_object strm);
+	void (*clear_input)(cl_object strm);
+	void (*clear_output)(cl_object strm);
+	void (*finish_output)(cl_object strm);
+	void (*force_output)(cl_object strm);
+
+	int (*input_p)(cl_object strm);
+	int (*output_p)(cl_object strm);
+	int (*interactive_p)(cl_object strm);
+	cl_object (*element_type)(cl_object strm);
+
+	cl_object (*length)(cl_object strm);
+	cl_object (*get_position)(cl_object strm);
+	cl_object (*set_position)(cl_object strm, cl_object pos);
+	int (*column)(cl_object strm);
+
+	cl_object (*close)(cl_object strm);
+};
+
+enum {
+	ECL_STREAM_BINARY = 0,
+	ECL_STREAM_FORMAT = 0xF,
+#ifndef ECL_UNICODE
+	ECL_STREAM_DEFAULT_FORMAT = 1,
+#else
+	ECL_STREAM_DEFAULT_FORMAT = 2,
+	ECL_STREAM_ISO_8859_1 = 1,
+	ECL_STREAM_LATIN_1 = 1,
+	ECL_STREAM_UTF_8 = 2,
+	ECL_STREAM_UCS_2 = 3,
+	ECL_STREAM_UCS_2LE = 5 + 128,
+	ECL_STREAM_UCS_2BE = 5,
+	ECL_STREAM_UCS_4 = 6,
+	ECL_STREAM_UCS_4LE = 7 + 128,
+	ECL_STREAM_UCS_4BE = 7,
+	ECL_STREAM_USER_FORMAT = 8,
+	ECL_STREAM_US_ASCII = 10,
+#endif
+	ECL_STREAM_CR = 16,
+	ECL_STREAM_LF = 32,
+	ECL_STREAM_SIGNED_BYTES = 64,
+	ECL_STREAM_LITTLE_ENDIAN = 128,
+	ECL_STREAM_C_STREAM = 256,
+	ECL_STREAM_MIGHT_SEEK = 512,
+	ECL_STREAM_CLOSE_COMPONENTS = 1024
+};
+
+typedef int (*cl_eformat_encoder)(cl_object stream, unsigned char *buffer, int c);
+typedef cl_index (*cl_eformat_read_byte8)(cl_object object, unsigned char *buffer, cl_index n);
+typedef int (*cl_eformat_decoder)(cl_object stream, cl_eformat_read_byte8 read_byte8, cl_object source);
+
+#define ECL_ANSI_STREAM_P(o) \
+        (IMMEDIATE(o) == 0 && ((o)->d.t == t_stream))
+#define ECL_ANSI_STREAM_TYPE_P(o,m) \
+        (IMMEDIATE(o) == 0 && ((o)->d.t == t_stream) && ((o)->stream.mode == (m)))
+
 struct ecl_stream {
-	HEADER4(mode,closed,char_stream_p,signed_bytes);
-				/*  stream mode of enum smmode  */
-				/*  stream element type  */
-	void	*file;		/*  file pointer  */
+	HEADER2(mode,closed);	/*  stream mode of enum smmode  */
+				/*  closed stream?  */
+	struct ecl_file_ops *ops; /*  dispatch table  */
+        union {
+                FILE *stream;		/* ANSI C streams */
+                cl_fixnum descriptor;	/* POSIX files */
+        } file;
 	cl_object object0;	/*  some object  */
 	cl_object object1;	/*  some object */
+	cl_object byte_stack;	/*  buffer for unread bytes  */
+	cl_fixnum last_char;	/*  last character read  */
+	cl_fixnum last_code[2];	/*  actual composition of last character  */
 	cl_fixnum int0;		/*  some int  */
 	cl_fixnum int1;		/*  some int  */
-	char	*buffer;	/*  file buffer  */
 	cl_index byte_size;	/*  size of byte in binary streams  */
-	unsigned char bit_buffer;
-	uint8_t bits_left;
-	int8_t buffer_state;	/* 0: unknown, 1: reading, -1: writing */
-	uint8_t header;		/* number of significant bits in the last byte */
-	int8_t last_op;		/* 0: unknown, 1: reading, -1: writing */
+	cl_fixnum last_op;	/*  0: unknown, 1: reading, -1: writing */
+	char *buffer;		/*  buffer for FILE  */
+	cl_object format;	/*  external format  */
+	cl_eformat_encoder encoder;
+	cl_eformat_decoder decoder;
+	cl_object format_table;
+	int flags;		/*  character table, flags, etc  */
 };
 
 struct ecl_random {
@@ -506,26 +675,23 @@ enum ecl_chattrib {		/*  character attribute  */
 
 struct ecl_readtable_entry {		/*  read table entry  */
 	enum ecl_chattrib syntax_type;	/*  character attribute  */
-	cl_object macro;		/*  macro function  */
-	cl_object *dispatch_table;	/*  pointer to the  */
-					/*  dispatch table  */
-					/*  NULL for  */
-					/*  non-dispatching  */
-					/*  macro character, or  */
-					/*  non-macro character  */
+	cl_object dispatch;		/*  a macro, a hash or NIL  */
 };
 
 enum ecl_readtable_case {
 	ecl_case_upcase,
 	ecl_case_downcase,
 	ecl_case_invert,
-	ecl_case_preserve,
+	ecl_case_preserve
 };
 
-struct ecl_readtable {			/*  read table  */
-	HEADER;
+struct ecl_readtable {		/*  read table  */
+	HEADER1(locked);
 	enum ecl_readtable_case read_case; /*  readtable-case  */
 	struct ecl_readtable_entry *table; /*  read table itself  */
+#ifdef ECL_UNICODE
+	cl_object hash;		/* hash for values outside base-char range */
+#endif
 };
 
 struct ecl_pathname {
@@ -554,39 +720,58 @@ struct ecl_codeblock {
 	cl_object links;		/*  list of symbols with linking calls  */
 	cl_index cfuns_size;		/*  number of functions defined  */
 	const struct ecl_cfun *cfuns;
+        cl_object source;		/*  common debug information for this block  */
 };
 
 struct ecl_bytecodes {
 	HEADER;
 	cl_object name;		/*  function name  */
-	cl_object specials;	/*  list of special variables  */
 	cl_object definition;	/*  function definition in list form  */
+	cl_objectfn entry;	/*  entry address (must match the position of
+                                 *  the equivalent field in cfun) */
 	cl_index code_size;	/*  number of bytecodes  */
 	cl_index data_size;	/*  number of constants  */
 	char *code;		/*  the intermediate language  */
 	cl_object *data;	/*  non-inmediate constants used in the code  */
 	cl_object file;		/*  file where it was defined...  */
-	cl_index file_position;	/*  and where it was created  */
+	cl_object file_position;/*  and where it was created  */
 };
 
 struct ecl_bclosure {
 	HEADER;
 	cl_object code;
 	cl_object lex;
+	cl_objectfn entry;	/*  entry address  */
 };
 
 struct ecl_cfun {		/*  compiled function header  */
 	HEADER1(narg);
 	cl_object name;		/*  compiled function name  */
-	cl_objectfn entry;	/*  entry address  */
 	cl_object block;	/*  descriptor of C code block for GC  */
+	cl_objectfn entry;	/*  entry address  */
+	cl_object file;		/*  file where it was defined...  */
+	cl_object file_position;/*  and where it was created  */
+};
+
+struct ecl_cfunfixed {		/*  compiled function header  */
+	HEADER1(narg);
+	cl_object name;		/*  compiled function name  */
+	cl_object block;	/*  descriptor of C code block for GC  */
+	cl_objectfn entry;	/*  entry address (must match the position of
+                                 *  the equivalent field in cfun) */
+	cl_objectfn_fixed entry_fixed;	/*  entry address  */
+	cl_object file;		/*  file where it was defined...  */
+	cl_object file_position;/*  and where it was created  */
 };
 
 struct ecl_cclosure {		/*  compiled closure header  */
 	HEADER;
 	cl_object env;		/*  environment  */
-	cl_objectfn entry;	/*  entry address  */
 	cl_object block;	/*  descriptor of C code block for GC  */
+	cl_objectfn entry;	/*  entry address (must match the position of
+                                 *  the equivalent field in cfun) */
+	cl_object file;		/*  file where it was defined...  */
+	cl_object file_position;/*  and where it was created  */
 };
 
 #define ECL_FFICALL_LIMIT 256
@@ -602,6 +787,26 @@ enum ecl_ffi_tag {
 	ECL_FFI_UNSIGNED_INT,
 	ECL_FFI_LONG,
 	ECL_FFI_UNSIGNED_LONG,
+#ifdef ecl_uint8_t
+        ECL_FFI_INT8_T,
+        ECL_FFI_UINT8_T,
+#endif
+#ifdef ecl_uint16_t
+        ECL_FFI_INT16_T,
+        ECL_FFI_UINT16_T,
+#endif
+#ifdef ecl_uint32_t
+        ECL_FFI_INT32_T,
+        ECL_FFI_UINT32_T,
+#endif
+#ifdef ecl_uint64_t
+        ECL_FFI_INT64_T,
+        ECL_FFI_UINT64_T,
+#endif
+#ifdef ecl_long_long_t
+        ECL_FFI_LONG_LONG,
+        ECL_FFI_UNSIGNED_LONG_LONG,
+#endif
 	ECL_FFI_POINTER_VOID,
 	ECL_FFI_CSTRING,
 	ECL_FFI_OBJECT,
@@ -621,6 +826,27 @@ union ecl_ffi_values {
 	unsigned short us;
 	long l;
 	unsigned long ul;
+#ifdef ecl_uint8_t
+        ecl_int8_t i8;
+        ecl_uint8_t u8;
+#endif
+#ifdef ecl_uint16_t
+        ecl_int16_t i16;
+        ecl_uint16_t u16;
+#endif
+#ifdef ecl_uint32_t
+        ecl_int32_t i32;
+        ecl_uint32_t u32;
+#endif
+#ifdef ecl_uint64_t
+        ecl_int64_t i64;
+        ecl_uint64_t u64;
+#endif
+#ifdef ecl_long_long_t
+        ecl_long_long_t ll;
+        ecl_ulong_long_t ull;
+        unsigned long l2[2];
+#endif
 	void *pv;
 	char *pc;
 	cl_object o;
@@ -642,9 +868,15 @@ struct ecl_foreign {		/*  user defined datatype  */
 
 struct ecl_stack_frame {
 	HEADER;
-	cl_object *bottom;	/*  Bottom part  */
-	cl_object *top;		/*  Top part  */
 	cl_object *stack;	/*  Is this relative to the lisp stack?  */
+	cl_object *base;	/*  Start of frame  */
+        cl_index size;		/*  Number of arguments  */
+	struct cl_env_struct *env;
+};
+
+struct ecl_weak_pointer {	/*  weak pointer to value  */
+	HEADER;
+	cl_object value;
 };
 
 /*
@@ -663,10 +895,11 @@ struct ecl_process {
 	pthread_t thread;
 	struct cl_env_struct *env;
 	cl_object interrupt;
-	/*void *stack;*/
-	/*cl_index stack_size;*/
-	void *altstack;
-	cl_index altstack_size;
+	cl_object initial_bindings;
+        cl_object parent;
+	cl_object exit_lock;
+	int trap_fpe_bits;
+        cl_object exit_values;
 };
 
 struct ecl_lock {
@@ -682,7 +915,11 @@ struct ecl_condition_variable {
         pthread_cond_t cv;
 };
 
-#endif
+struct ecl_semaphore {
+        HEADER;
+        void *handle;
+};
+#endif /* ECL_THREADS */
 
 #ifdef CLOS
 #define CLASS_OF(x)		(x)->instance.clas
@@ -700,6 +937,7 @@ struct ecl_instance {		/*  instance header  */
 	HEADER1(isgf);
 	cl_index length;	/*  instance length  */
 	cl_object clas;		/*  instance class  */
+	cl_objectfn entry;	/*  entry address  */
 	cl_object sig;		/*  generation signature  */
 	cl_object *slots;	/*  instance slots  */
 };
@@ -736,6 +974,7 @@ union cl_lispunion {
 	struct ecl_bytecodes	bytecodes; 	/*  bytecompiled function / code */
 	struct ecl_bclosure	bclosure; 	/*  bytecompiled closure */
 	struct ecl_cfun		cfun;		/*  compiled function  */
+	struct ecl_cfunfixed	cfunfixed;	/*  compiled function  */
 	struct ecl_cclosure	cclosure; 	/*  compiled closure  */
 
 	struct ecl_dummy	d;		/*  dummy  */
@@ -748,10 +987,14 @@ union cl_lispunion {
 	struct ecl_process	process; 	/*  process  */
 	struct ecl_lock		lock; 		/*  lock  */
         struct ecl_condition_variable condition_variable; /*  condition-variable */
+#endif
+#ifdef ECL_SEMAPHORES
+        struct ecl_semaphore	semaphore;	/*  semaphore  */
 #endif
 	struct ecl_codeblock	cblock;		/*  codeblock  */
 	struct ecl_foreign	foreign; 	/*  user defined data type */
 	struct ecl_stack_frame	frame;		/*  stack frame  */
+	struct ecl_weak_pointer weak;		/*  weak pointers  */
 };
 
 /*
